::Cultists <- {}

DoIncludeScript("Cultists/spell_data.nut", this)

::Cultists.Cache <- {StunTrigger = {}, BurnTrigger = {}} // Cache triggers for stunning and burning players
::Cultists.PostponeFunctions <- [] // Execute code with delay, primarily used by spells

// Debug function for reloading spells
::Cultists.ReloadSpells <- function()
{
    DoIncludeScript("Cultists/spell_data.nut", this)
    foreach (index, func in ::Cultists.Spells) {::Cultists.Spells[index] <- func.bindenv(::Cultists.SpellUtil) }
}

// Utility functions to be utilized by spells for quicker workflow
::Cultists.SpellUtil <- {
    // Execute code after some time - primarily used to delay the the actual spell effects from the summoning particles
    ExecuteWithDelay = function(delay, func)
    { ::Cultists.PostponeFunctions.push([Time() + delay, func, this]) },

    // A shorthand for spawning entities
    AddEntity = function(classname, where, team = 0, custom_keyvalues = {})
    {
        local keyvalues = {origin = where, TeamNum = team}
        foreach (key, value in custom_keyvalues) { keyvalues[key] = value }

        return SpawnEntityFromTable(classname, keyvalues)
    },

    // Create a script scope on an entity with the same variables as the current scope
    AddScriptToEntity = function(entity = null)
    {
        entity.ValidateScriptScope()

        foreach(key, value in this)
        { entity.GetScriptScope()[key] <- value }

        return entity.GetScriptScope()
    },

    // A shorthand for emitting particles
    DoParticle = function(particle_system, where, lifetime = 0.5, team = 0)
    {
        if (team > 0)
        { particle_system = (team == 2) ? particle_system + "_red" : particle_system + "_blue"}
        
        if (lifetime != null && lifetime <= 0.0)
        {
            DispatchParticleEffect(particle_system, where, Vector())
            return null
        }

        local system = SpawnEntityFromTable("info_particle_system", {origin = where, effect_name = particle_system})
        EntFireByHandle(system, "Start", "", 0.0, null, null)

        if (lifetime != null)
        {
            EntFireByHandle(system, "Stop", "", lifetime, null, null)
            EntFireByHandle(system, "Kill", "", lifetime + 0.02, null, null)
        }

        return system
    },

    // A shorthand for playing sounds from files
    DoSoundFile = function(sound_file, where, volume = 1.0, pitch = 1.0)
    { PrecacheSound(sound_file); EmitSoundEx({sound_name = sound_file, origin = where, volume = volume, pitch = pitch * 100}) },

    // A shorthand for playing sounds from soundscripts
    DoSoundScript = function(sound_script, where, volume = 1.0, pitch = 1.0)
    { PrecacheScriptSound(sound_script); EmitSoundEx({sound_name = sound_script, origin = where, volume = volume, pitch = pitch * 100}) },

    // A shorthand for "reliable" tracing - more expensive, more "accurate" in certain cases
    FitTrace = function(start, end, bounds_min = Vector(-8, -8, -8), bounds_max = Vector(8, 8, 8))
    {
        local ceiling_vector = start
        ceiling_vector.z = end.z

        local ceiling_trace = {
            start = start,
            end = ceiling_vector,
            hullmin = bounds_min,
            hullmax = bounds_max,
            mask = (Constants.FContents.CONTENTS_SOLID | Constants.FContents.CONTENTS_MOVEABLE | Constants.FContents.CONTENTS_WINDOW | Constants.FContents.CONTENTS_GRATE | Constants.FContents.CONTENTS_PLAYERCLIP),
        }
        TraceHull(ceiling_trace)

        local ceiling_vector = ceiling_trace.endpos

        local xy_trace = {
            start = ceiling_vector,
            end = end
            hullmin = bounds_min,
            hullmax = bounds_max,
            mask = (Constants.FContents.CONTENTS_SOLID | Constants.FContents.CONTENTS_MOVEABLE | Constants.FContents.CONTENTS_WINDOW | Constants.FContents.CONTENTS_GRATE | Constants.FContents.CONTENTS_PLAYERCLIP),
        }
        TraceHull(xy_trace)

        return xy_trace
    },

    // Summon the casting visuals
    ManifestVisuals = function(where, team = 0)
    {
        DoParticle("spell_teleport", where, 1.25, team)
        DoParticle("spell_cast_wheel", where, 0.0, team)
        DoSoundFile("ambient/levels/citadel/portal_beam_shoot5.wav", where)
        DoSoundFile("misc/halloween/spell_meteor_cast.wav", where, 0.9)
    },

    // A shorthand for varying aerial spawn positions
    SpawnVariance = function(start_position, xy_variance = 200, z_offset = 128)
    {
        local desired_position = Vector(start_position.x + RandomInt(-(xy_variance / 2), xy_variance / 2), start_position.y + RandomInt(-(xy_variance / 2), xy_variance / 2), start_position.z + z_offset)
        return FitTrace(start_position, desired_position).endpos
    },

    // Ignite a player
    BurnPlayer = function(player, duration = 10)
    {
        duration = duration.tointeger()
        local trigger = null

        if (duration in ::Cultists.Cache.BurnTrigger == false || ::Cultists.Cache.BurnTrigger[duration].IsValid() == false)
        {
            trigger = SpawnEntityFromTable("trigger_ignite", {
                burn_duration = 10,
                spawnflags = 1,
            })
            trigger.SetSolid(Constants.ESolidType.SOLID_NONE)

            ::Cultists.Cache.BurnTrigger[duration] <- trigger
        } else {
            trigger = ::Cultists.Cache.BurnTrigger[duration]
        }
        
        EntFireByHandle(trigger, "StartTouch", "", 0.00, player, player)
        EntFireByHandle(trigger, "EndTouch", "", 0.03, player, player)
    },

    // Scare a player
    ScarePlayer = function(player, duration = 5)
    {
        duration = duration.tointeger()
        local trigger = null

        if (duration in ::Cultists.Cache.StunTrigger == false || ::Cultists.Cache.StunTrigger[duration].IsValid() == false)
        {
            trigger = SpawnEntityFromTable("trigger_stun", {
                stun_type = 2,
                stun_effects = true,
                stun_duration = duration,
                trigger_delay = 0,
                spawnflags = 1,
            })
            trigger.SetSolid(Constants.ESolidType.SOLID_NONE)

            ::Cultists.Cache.StunTrigger[duration] <- trigger
        } else {
            trigger = ::Cultists.Cache.StunTrigger[duration]
        }
        
        EntFireByHandle(trigger, "EndTouch", "", 0.00, player, player)
        EntFireByHandle(trigger, "Kill", "", 0.03, player, player) // Lazy fix for an issue: players gain immunity to trigger_stuns if they trigger them indirectly??? (e.g. StartTouch/EndTouch)
    },
}
::Cultists.ReloadSpells()

function OnPostSpawn()
{ AddThinkToEnt(self, "Tick") }

function Tick()
{
    if (::Cultists.PostponeFunctions.len() == 0)
    { return -1 }

    foreach (index, func in ::Cultists.PostponeFunctions)
    {
        if (func[0] > Time()) { continue }
        func[1].call(func[2])
        ::Cultists.PostponeFunctions.remove(index)
    }

    return -1
}