// Use this constant in conjunction with ExecuteWithDelay() for a consistent delay between the spell manifesting visuals and the actual spell effects
const SPELL_MANIFEST_DELAY = 0.4

::Cultists.Spells <- {
    Jar = function(caster)
    {
        local spawn_position = SpawnVariance(caster.GetOrigin())
        local team = caster.GetTeam()

        ManifestVisuals(spawn_position, team)

        ExecuteWithDelay(SPELL_MANIFEST_DELAY, function() {
            local entity_types = ["", "_milk", "_gas"]
            AddEntity("tf_projectile_jar" + entity_types[RandomInt(0, 2)], spawn_position, team).SetAbsAngles(QAngle(-50, RandomInt(0,359), 0))
        })
    },

    // Required by the Armageddon spell
    Fireball = function(caster, custom_position = null, custom_team = null)
    {
        local spawn_position = custom_position ? custom_position : SpawnVariance(caster.GetOrigin())
        local team = custom_team ? custom_team : caster.GetTeam()

        ManifestVisuals(spawn_position, team)

        ExecuteWithDelay(SPELL_MANIFEST_DELAY, function() {
            local projectile = AddEntity("tf_projectile_spellfireball", spawn_position, 2)
            projectile.SetMoveType(Constants.EMoveType.MOVETYPE_NONE, Constants.EMoveCollide.MOVECOLLIDE_DEFAULT)
            projectile.SetSolid(Constants.ESolidType.SOLID_NONE)
            projectile.AddEFlags(Constants.FEntityEFlags.EFL_NO_THINK_FUNCTION) // Note to self: Does this even do anything?

            // Add custom projectile behaviour
            AddScriptToEntity(projectile)
            projectile.GetScriptScope().GravityStrength <- 0.17
            projectile.GetScriptScope().Tick <- function()
            {
                local current_velocity = self.GetAbsVelocity()
                local next_position = self.GetOrigin() + current_velocity

                local trace_line = {
                    start = self.GetOrigin(),
                    end = next_position,
                    ignore = self,
                    mask = (
                        Constants.FContents.CONTENTS_SOLID
                        | Constants.FContents.CONTENTS_GRATE
                        | Constants.FContents.CONTENTS_WINDOW
                        | Constants.FContents.CONTENTS_DEBRIS
                        | Constants.FContents.CONTENTS_MOVEABLE
                        | Constants.FContents.CONTENTS_DETAIL
                        | Constants.FContents.CONTENTS_PLAYERCLIP
                        | Constants.FContents.CONTENTS_HITBOX
                    ),
                }
                TraceLineEx(trace_line)

                // Detect if we hit something
                if (trace_line.fraction < 1.0)
                {
                    local hit_position = trace_line.endpos
                    local targets = []

                    local target_these_classes = ["player", "obj_sentrygun", "obj_dispenser", "obj_teleporter"]
                    local on_this_team = (team == 2) ? 3 : 2
                    local within_this_radius = 224
                    local no_higher_than = 48

                    foreach (classname in target_these_classes)
                    {
                        local what = null
                        while (what = Entities.FindByClassnameWithin(what, classname, hit_position, within_this_radius))
                        {
                            if (what.GetTeam() != on_this_team || what.GetOrigin().z - hit_position.z > no_higher_than)
                            { continue }
                            
                            // Don't hurt things behind cover
                            local trace_line = {
                                start = hit_position,
                                end = what.GetOrigin(),
                                mask = (
                                    Constants.FContents.CONTENTS_SOLID
                                    | Constants.FContents.CONTENTS_GRATE
                                    | Constants.FContents.CONTENTS_WINDOW
                                    | Constants.FContents.CONTENTS_DEBRIS
                                    | Constants.FContents.CONTENTS_MOVEABLE
                                    | Constants.FContents.CONTENTS_DETAIL
                                    | Constants.FContents.CONTENTS_PLAYERCLIP
                                ),
                            }
                            TraceLineEx(trace_line)

                            if (trace_line.fraction == 1.0)
                            {
                                // Damage things if all checks pass
                                if (what.GetClassname() == "player")
                                {
                                    what.TakeDamageEx(self, null, null, Vector(), hit_position, 30.0, Constants.FDmgType.DMG_BURN)
                                    BurnPlayer(what)
                                } else {
                                    // Source jank causes buildings to crash the game when dying to TakeDamageEx
                                    EntFireByHandle(what, "RemoveHealth", "30", 0.00, null, null)
                                }
                            }
                        }
                    }

                    // Fire circle effects
                    DoParticle("heavy_ring_of_fire", hit_position)
                    DoSoundScript("Weapon_DragonsFury.Nearmiss", hit_position)

                    self.Kill()
                    return -1
                }

                self.SetOrigin(next_position)
                
                // Fall down to the ground
                local gravity_multiplier = Convars.GetFloat("sv_gravity") / 800 // Lazy way of doing this but it works
                local z_velocity = current_velocity.z - (GravityStrength * gravity_multiplier)
                current_velocity.z = z_velocity

                self.SetAbsVelocity(current_velocity + self.GetBaseVelocity())

                return -1
            }
            
            AddThinkToEnt(projectile, "Tick")
        })
    },

    // Spell dependencies: Fireball
    Armageddon = function(caster)
    {
        local formations = [
            Vector(1, -1, 0),
            Vector(1, 1, 0),
            Vector(-1, 0, 0),
        ]
        local range_minimum = 128.0
        local range_maximum = 256.0
        
        local team = caster.GetTeam()

        foreach (index, formation in formations)
        {

            formation = RotatePosition(Vector(0, 0, 0), caster.GetAbsAngles(), formation)
            formation = formation * RandomFloat(range_minimum, range_maximum)
            formation.z = 128.0

            local spawn_position = caster.GetOrigin() + formation

            local fit_trace = FitTrace(caster.GetOrigin(), spawn_position)

            spawn_position = fit_trace.endpos

            ExecuteWithDelay(0.2 * index, function() {
                Cultists.Spells.Fireball(caster, spawn_position, team)
            })
        }
    },

    // Required by the BombPack spell
    Bomb = function(caster, custom_position = null)
    {
        local start_position = custom_position ? custom_position : SpawnVariance(caster.GetOrigin(), 256, 24)
        local final_position = Vector(start_position.x, start_position.y, start_position.z - 9999)
        local team = caster.GetTeam()

        local hull_trace = {
            start = start_position,
            end = final_position,
            hullmin = Vector(-20, -20, 0),
            hullmax = Vector(20, 20, 32),
            mask = (
                Constants.FContents.CONTENTS_SOLID
                | Constants.FContents.CONTENTS_GRATE
                | Constants.FContents.CONTENTS_WINDOW
                | Constants.FContents.CONTENTS_DEBRIS
                | Constants.FContents.CONTENTS_MOVEABLE
                | Constants.FContents.CONTENTS_DETAIL
                | Constants.FContents.CONTENTS_PLAYERCLIP
            ),
        }
        TraceHull(hull_trace)

        local spawn_position = hull_trace.endpos

        ManifestVisuals(spawn_position + Vector(0, 0, 8), team)

        ExecuteWithDelay(SPELL_MANIFEST_DELAY, function() {
            DoParticle("ghost_appearation", spawn_position)
            DoSoundScript("Halloween.PumpkinDrop", spawn_position)

            local bomb = Entities.CreateByClassname("tf_generic_bomb")
            bomb.SetModelSimple("models/props_halloween/pumpkin_explode_teamcolor.mdl")
            bomb.SetSkin(team - 1)
            bomb.SetOrigin(spawn_position)
            bomb.SetAbsAngles(QAngle(0, RandomInt(0, 359), 0))
            bomb.KeyValueFromString("sound", "Halloween.PumpkinExplode")
            bomb.KeyValueFromString("explode_particle", "pumpkin_explode")
            bomb.KeyValueFromInt("damage", 40)
            bomb.KeyValueFromInt("radius", 100)
            bomb.KeyValueFromInt("friendly_fire", 1)
            Entities.DispatchSpawn(bomb)

            // Be nice and don't get players stuck
            local dont_collide = ["player", "tf_robot_destruction_robot"]
            local entity = null
            foreach (classname in dont_collide)
            {
                entity = Entities.FindByClassnameWithin(null, classname, bomb.GetCenter(), 64.0)

                if (entity != null)
                { break }
            }

            if (entity != null)
                { 
                    bomb.SetSolid(Constants.ESolidType.SOLID_NONE)
                    bomb.KeyValueFromInt("rendermode", Constants.ERenderMode.kRenderTransAlpha)
                    bomb.KeyValueFromInt("renderamt", 130)
                } else {
                    bomb.SetSolid(Constants.ESolidType.SOLID_BBOX)
                }

            // Bomb scripts
            AddScriptToEntity(bomb)
            bomb.GetScriptScope().ExpireTime <- Time() + 60.0
            bomb.GetScriptScope().KeepCollisionsOff <- (entity != null)
            bomb.GetScriptScope().Tick <- function()
            {
                if (KeepCollisionsOff == true)
                {
                    local dont_collide = ["player", "tf_robot_destruction_robot"]
                    local entity = null
                    foreach (classname in dont_collide)
                    {
                        entity = Entities.FindByClassnameWithin(null, classname, bomb.GetCenter(), 64.0)

                        if (entity != null)
                        { break }
                    }
                    
                    if (entity == null)
                    {
                        bomb.SetSolid(Constants.ESolidType.SOLID_BBOX)
                        bomb.KeyValueFromInt("rendermode", Constants.ERenderMode.kRenderNormal)
                        bomb.KeyValueFromInt("renderamt", 255)
                        KeepCollisionsOff == false
                    }
                }

                if (Time() > ExpireTime)
                {
                    DoParticle("ghost_appearation", self.GetOrigin())
                    DoSoundScript("Halloween.spell_spawn_boss_disappear", self.GetOrigin())
                    self.Kill()
                }

                return 0.5
            }
            AddThinkToEnt(bomb, "Tick")
        })
    },

    // Spell dependencies: Bomb
    BombPack = function(caster)
    {
        local formations = [
            Vector(0, 1, 0),
            Vector(0, -1, 0),
            Vector(-1, 0, 0),
        ]
        local range_minimum = 64.0
        local range_maximum = 128.0
        
        foreach (formation in formations)
        {

            formation = RotatePosition(Vector(0, 0, 0), caster.GetAbsAngles(), formation)
            formation = formation * RandomFloat(range_minimum, range_maximum)
            formation.z = 128.0

            local spawn_position = caster.GetOrigin() + formation

            local fit_trace = FitTrace(caster.GetOrigin(), spawn_position)

            spawn_position = fit_trace.endpos

            Cultists.Spells.Bomb(caster, spawn_position)
        }
    }

    Jumpscare = function(caster)
    {
        local angles = caster.GetAbsAngles()
        local spawn_position = caster.GetOrigin() + (angles.Forward() * -64)
        local team = caster.GetTeam()

        local hull_trace = {
            start = caster.GetOrigin(),
            end = spawn_position,
            hullmin = Vector(-28, -28, 0),
            hullmax = Vector(28, 28, 92),
            mask = (Constants.FContents.CONTENTS_SOLID | Constants.FContents.CONTENTS_MOVEABLE | Constants.FContents.CONTENTS_WINDOW | Constants.FContents.CONTENTS_GRATE | Constants.FContents.CONTENTS_PLAYERCLIP),
        }
        TraceHull(hull_trace)
        spawn_position = hull_trace.endpos
        
        ManifestVisuals(spawn_position, team)

        ExecuteWithDelay(SPELL_MANIFEST_DELAY, function()
        {
            DoParticle("ghost_appearation", spawn_position + Vector(0, 0, 60))
            DoSoundScript("Halloween.Haunted", spawn_position, 8)

            ExecuteWithDelay(0.1, function()
            {
                local which_model = (team == 2) ? "models/props_halloween/ghost_no_hat_red.mdl" : "models/props_halloween/ghost_no_hat.mdl"
                local ghost_prop = SpawnEntityFromTable("prop_dynamic", {
                    origin = spawn_position + Vector(0, 0, 5),
                    angles = angles.ToKVString(),
                    model = which_model,
                    solid = Constants.ESolidType.SOLID_NONE,
                    TeamNum = team,
                    DefaultAnim = "idle",
                })

                AddScriptToEntity(ghost_prop)
                ghost_prop.GetScriptScope().DisappearTime <- Time() + 1.6
                ghost_prop.GetScriptScope().Tick <- function ()
                {
                    local baby = null
                    while (baby = Entities.FindByClassnameWithin(baby, "player", spawn_position, 256))
                    {
                        if (baby.GetTeam() == self.GetTeam()) { continue }
                        
                        // Trace a line to make sure to not scare players behind cover
                        local trace_line = {
                            start = self.GetOrigin(),
                            end = baby.GetOrigin(),
                            mask = (Constants.FContents.CONTENTS_SOLID | Constants.FContents.CONTENTS_MOVEABLE),
                        }
                        TraceLineEx(trace_line)

                        // You can see me! Boo!
                        if (trace_line.fraction == 1.0)
                        {
                            ScarePlayer(baby)
                        }
                    }

                    if (Time() > DisappearTime)
                    {
                        DoParticle("ghost_appearation", self.GetOrigin() + Vector(0, 0, 60))
                        self.Kill()
                    }

                    return 0.4
                }
                AddThinkToEnt(ghost_prop, "Tick")
                
                DoSoundScript("Halloween.GhostBoo", spawn_position, 8)
            })
        })
    },
    
    Skeletons = function(caster)
    {
        local team = caster.GetTeam()

        local skeleton_health = 120.0

        for (local n = -1; n < 2; n = n + 2)
        {
            local spawn_position = caster.GetOrigin() + RotatePosition(Vector(0, 0, 0), caster.GetAbsAngles(), Vector(0, n * 64, 0)) + Vector(0, 0, 48)

            local hull_trace = {
                start = caster.GetOrigin() + Vector(0, 0, 48),
                end = spawn_position,
                hullmin = Vector(-28, -14, -45),
                hullmax = Vector(28, 14, 45),
                mask = (Constants.FContents.CONTENTS_SOLID | Constants.FContents.CONTENTS_MOVEABLE | Constants.FContents.CONTENTS_WINDOW | Constants.FContents.CONTENTS_GRATE | Constants.FContents.CONTENTS_PLAYERCLIP),
            }
            TraceHull(hull_trace)

            // Glue the skeletons to the ground so their spawn animations look less off
            local ground_trace = {
                start = hull_trace.endpos,
                end = Vector(hull_trace.endpos.x, hull_trace.endpos.y, hull_trace.endpos.z - 9999),
                mask = (Constants.FContents.CONTENTS_SOLID | Constants.FContents.CONTENTS_MOVEABLE | Constants.FContents.CONTENTS_WINDOW | Constants.FContents.CONTENTS_GRATE | Constants.FContents.CONTENTS_PLAYERCLIP),
            }
            TraceLineEx(ground_trace)

            spawn_position = ground_trace.endpos

            ManifestVisuals(spawn_position, team)
            
            ExecuteWithDelay(SPELL_MANIFEST_DELAY, function(){
                local skeleton = AddEntity("tf_zombie", spawn_position, team)
                skeleton.SetSkin(team - 2)
                skeleton.SetMaxHealth(skeleton_health)
                skeleton.SetHealth(skeleton_health)
            })
        }
    },
}
